@@ -1,25 +1,23 @@
-# coding: utf-8
-from __future__ import absolute_import, division, print_function, unicode_literals
-
 import os
 import shlex
 import sys
 import tempfile
-import warnings
+from typing import Any, BinaryIO, Optional, Tuple, cast
 
-from click import Command
-from click.utils import safecall
+import click
+from click.utils import LazyFile, safecall
 from pip._internal.commands import create_command
 from pip._internal.req.constructors import install_req_from_line
 from pip._internal.utils.misc import redact_auth_from_url
+from pip._vendor.pep517 import meta
 
-from .. import click
 from .._compat import parse_requirements
 from ..cache import DependencyCache
 from ..exceptions import PipToolsError
 from ..locations import CACHE_DIR
 from ..logging import log
 from ..repositories import LocalRequirementsRepository, PyPIRepository
+from ..repositories.base import BaseRepository
 from ..resolver import Resolver
 from ..utils import UNSAFE_PACKAGES, dedup, is_pinned_requirement, key_from_ireq
 from ..writer import OutputWriter
@@ -28,7 +26,7 @@ DEFAULT_REQUIREMENTS_FILE = "requirements.in"
 DEFAULT_REQUIREMENTS_OUTPUT_FILE = "requirements.txt"
 
 
-def _get_default_option(option_name):
+def _get_default_option(option_name: str) -> Any:
     """
     Get default value of the pip's option (including option from pip.conf)
     by a given option name.
@@ -38,28 +36,7 @@ def _get_default_option(option_name):
     return getattr(default_values, option_name)
 
 
-class BaseCommand(Command):
-    _os_args = None
-
-    def parse_args(self, ctx, args):
-        """
-        Override base `parse_args` to store the argument part of `sys.argv`.
-        """
-        self._os_args = set(args)
-        return super(BaseCommand, self).parse_args(ctx, args)
-
-    def has_arg(self, arg_name):
-        """
-        Detect whether a given arg name (including negative counterparts
-        to the arg, e.g. --no-arg) is present in the argument part of `sys.argv`.
-        """
-        command_options = {option.name: option for option in self.params}
-        option = command_options[arg_name]
-        args = set(option.opts + option.secondary_opts)
-        return bool(self._os_args & args)
-
-
-@click.command(cls=BaseCommand)
+@click.command(context_settings={"help_option_names": ("-h", "--help")})
 @click.version_option()
 @click.pass_context
 @click.option("-v", "--verbose", count=True, help="Show more output")
@@ -88,7 +65,6 @@ class BaseCommand(Command):
     "--find-links",
     multiple=True,
     help="Look for archives in this directory or on this HTML page",
-    envvar="PIP_FIND_LINKS",
 )
 @click.option(
     "-i",
@@ -96,13 +72,9 @@ class BaseCommand(Command):
     help="Change index URL (defaults to {index_url})".format(
         index_url=redact_auth_from_url(_get_default_option("index_url"))
     ),
-    envvar="PIP_INDEX_URL",
 )
 @click.option(
-    "--extra-index-url",
-    multiple=True,
-    help="Add additional index URL to search",
-    envvar="PIP_EXTRA_INDEX_URL",
+    "--extra-index-url", multiple=True, help="Add additional index URL to search"
 )
 @click.option("--cert", help="Path to alternate CA bundle.")
 @click.option(
@@ -113,7 +85,6 @@ class BaseCommand(Command):
 @click.option(
     "--trusted-host",
     multiple=True,
-    envvar="PIP_TRUSTED_HOST",
     help="Mark this host as trusted, even though it does not have "
     "valid or any HTTPS.",
 )
@@ -123,12 +94,6 @@ class BaseCommand(Command):
     default=True,
     help="Add header to generated file",
 )
-@click.option(
-    "--index/--no-index",
-    is_flag=True,
-    default=True,
-    help="DEPRECATED: Add index URL to generated file",
-)
 @click.option(
     "--emit-trusted-host/--no-emit-trusted-host",
     is_flag=True,
@@ -168,11 +133,16 @@ class BaseCommand(Command):
     ),
 )
 @click.option(
-    "--allow-unsafe",
+    "--allow-unsafe/--no-allow-unsafe",
     is_flag=True,
     default=False,
-    help="Pin packages considered unsafe: {}".format(
-        ", ".join(sorted(UNSAFE_PACKAGES))
+    help=(
+        "Pin packages considered unsafe: {}.\n\n"
+        "WARNING: Future versions of pip-tools will enable this behavior by default. "
+        "Use --no-allow-unsafe to keep the old behavior. It is recommended to pass the "
+        "--allow-unsafe now to adapt to the upcoming change.".format(
+            ", ".join(sorted(UNSAFE_PACKAGES))
+        )
     ),
 )
 @click.option(
@@ -181,6 +151,15 @@ class BaseCommand(Command):
     default=False,
     help="Generate pip 8 style hashes in the resulting requirements file.",
 )
+@click.option(
+    "--reuse-hashes/--no-reuse-hashes",
+    is_flag=True,
+    default=True,
+    help=(
+        "Improve the speed of --generate-hashes by reusing the hashes from an "
+        "existing output file."
+    ),
+)
 @click.option(
     "--max-rounds",
     default=10,
@@ -205,12 +184,12 @@ class BaseCommand(Command):
     "--cache-dir",
     help="Store the cache data in DIRECTORY.",
     default=CACHE_DIR,
-    envvar="PIP_TOOLS_CACHE_DIR",
     show_default=True,
-    show_envvar=True,
     type=click.Path(file_okay=False, writable=True),
 )
-@click.option("--pip-args", help="Arguments to pass directly to the pip command.")
+@click.option(
+    "--pip-args", "pip_args_str", help="Arguments to pass directly to the pip command."
+)
 @click.option(
     "--emit-index-url/--no-emit-index-url",
     is_flag=True,
@@ -218,35 +197,35 @@ class BaseCommand(Command):
     help="Add index URL to generated file",
 )
 def cli(
-    ctx,
-    verbose,
-    quiet,
-    dry_run,
-    pre,
-    rebuild,
-    find_links,
-    index_url,
-    extra_index_url,
-    cert,
-    client_cert,
-    trusted_host,
-    header,
-    index,
-    emit_trusted_host,
-    annotate,
-    upgrade,
-    upgrade_packages,
-    output_file,
-    allow_unsafe,
-    generate_hashes,
-    src_files,
-    max_rounds,
-    build_isolation,
-    emit_find_links,
-    cache_dir,
-    pip_args,
-    emit_index_url,
-):
+    ctx: click.Context,
+    verbose: int,
+    quiet: int,
+    dry_run: bool,
+    pre: bool,
+    rebuild: bool,
+    find_links: Tuple[str],
+    index_url: str,
+    extra_index_url: Tuple[str],
+    cert: Optional[str],
+    client_cert: Optional[str],
+    trusted_host: Tuple[str],
+    header: bool,
+    emit_trusted_host: bool,
+    annotate: bool,
+    upgrade: bool,
+    upgrade_packages: Tuple[str],
+    output_file: Optional[LazyFile],
+    allow_unsafe: bool,
+    generate_hashes: bool,
+    reuse_hashes: bool,
+    src_files: Tuple[str],
+    max_rounds: int,
+    build_isolation: bool,
+    emit_find_links: bool,
+    cache_dir: str,
+    pip_args_str: Optional[str],
+    emit_index_url: bool,
+) -> None:
     """Compiles requirements.txt from requirements.in specs."""
     log.verbosity = verbose - quiet
 
@@ -268,8 +247,10 @@ def cli(
         if src_files == ("-",):
             raise click.BadParameter("--output-file is required if input is from stdin")
         # Use default requirements output file if there is a setup.py the source file
-        elif src_files == ("setup.py",):
-            file_name = DEFAULT_REQUIREMENTS_OUTPUT_FILE
+        elif os.path.basename(src_files[0]) == "setup.py":
+            file_name = os.path.join(
+                os.path.dirname(src_files[0]), DEFAULT_REQUIREMENTS_OUTPUT_FILE
+            )
         # An output file must be provided if there are multiple source files
         elif len(src_files) > 1:
             raise click.BadParameter(
@@ -283,49 +264,35 @@ def cli(
         output_file = click.open_file(file_name, "w+b", atomic=True, lazy=True)
 
         # Close the file at the end of the context execution
+        assert output_file is not None
         ctx.call_on_close(safecall(output_file.close_intelligently))
 
-    if cli.has_arg("index") and cli.has_arg("emit_index_url"):
-        raise click.BadParameter(
-            "--index/--no-index and --emit-index-url/--no-emit-index-url "
-            "are mutually exclusive."
-        )
-    elif cli.has_arg("index"):
-        warnings.warn(
-            "--index and --no-index are deprecated and will be removed "
-            "in future versions. Use --emit-index-url/--no-emit-index-url instead.",
-            category=FutureWarning,
-        )
-        emit_index_url = index
-
     ###
     # Setup
     ###
 
-    right_args = shlex.split(pip_args or "")
+    right_args = shlex.split(pip_args_str or "")
     pip_args = []
-    if find_links:
-        for link in find_links:
-            pip_args.extend(["-f", link])
+    for link in find_links:
+        pip_args.extend(["-f", link])
     if index_url:
         pip_args.extend(["-i", index_url])
-    if extra_index_url:
-        for extra_index in extra_index_url:
-            pip_args.extend(["--extra-index-url", extra_index])
+    for extra_index in extra_index_url:
+        pip_args.extend(["--extra-index-url", extra_index])
     if cert:
         pip_args.extend(["--cert", cert])
     if client_cert:
         pip_args.extend(["--client-cert", client_cert])
     if pre:
         pip_args.extend(["--pre"])
-    if trusted_host:
-        for host in trusted_host:
-            pip_args.extend(["--trusted-host", host])
+    for host in trusted_host:
+        pip_args.extend(["--trusted-host", host])
 
     if not build_isolation:
         pip_args.append("--no-build-isolation")
     pip_args.extend(right_args)
 
+    repository: BaseRepository
     repository = PyPIRepository(pip_args, cache_dir=cache_dir)
 
     # Parse all constraints coming from --upgrade-package/-P
@@ -358,7 +325,9 @@ def cli(
                 existing_pins_to_upgrade.add(key)
             else:
                 existing_pins[key] = ireq
-        repository = LocalRequirementsRepository(existing_pins, repository)
+        repository = LocalRequirementsRepository(
+            existing_pins, repository, reuse_hashes=reuse_hashes
+        )
 
     ###
     # Parsing/collecting initial requirements
@@ -367,23 +336,14 @@ def cli(
     constraints = []
     for src_file in src_files:
         is_setup_file = os.path.basename(src_file) == "setup.py"
-        if is_setup_file or src_file == "-":
+        if src_file == "-":
             # pip requires filenames and not files. Since we want to support
             # piping from stdin, we need to briefly save the input from stdin
             # to a temporary file and have pip read that.  also used for
             # reading requirements from install_requires in setup.py.
             tmpfile = tempfile.NamedTemporaryFile(mode="wt", delete=False)
-            if is_setup_file:
-                from distutils.core import run_setup
-
-                dist = run_setup(src_file)
-                tmpfile.write("\n".join(dist.install_requires))
-                comes_from = "{name} ({filename})".format(
-                    name=dist.get_name(), filename=src_file
-                )
-            else:
-                tmpfile.write(sys.stdin.read())
-                comes_from = "-r -"
+            tmpfile.write(sys.stdin.read())
+            comes_from = "-r -"
             tmpfile.flush()
             reqs = list(
                 parse_requirements(
@@ -396,6 +356,15 @@ def cli(
             for req in reqs:
                 req.comes_from = comes_from
             constraints.extend(reqs)
+        elif is_setup_file:
+            dist = meta.load(os.path.dirname(os.path.abspath(src_file)))
+            comes_from = f"{dist.metadata.get_all('Name')[0]} ({src_file})"
+            constraints.extend(
+                [
+                    install_req_from_line(req, comes_from=comes_from)
+                    for req in dist.requires or []
+                ]
+            )
         else:
             constraints.extend(
                 parse_requirements(
@@ -417,7 +386,13 @@ def cli(
 
     # Filter out pip environment markers which do not match (PEP496)
     constraints = [
-        req for req in constraints if req.markers is None or req.markers.evaluate()
+        req
+        for req in constraints
+        if req.markers is None
+        # We explicitly set extra=None to filter out optional requirements
+        # since evaluating an extra marker with no environment raises UndefinedEnvironmentName
+        # (see https://packaging.pypa.io/en/latest/markers.html#usage)
+        or req.markers.evaluate({"extra": None})
     ]
 
     log.debug("Using indexes:")
@@ -442,10 +417,7 @@ def cli(
             allow_unsafe=allow_unsafe,
         )
         results = resolver.resolve(max_rounds=max_rounds)
-        if generate_hashes:
-            hashes = resolver.resolve_hashes(results)
-        else:
-            hashes = None
+        hashes = resolver.resolve_hashes(results) if generate_hashes else None
     except PipToolsError as e:
         log.error(str(e))
         sys.exit(2)
@@ -457,8 +429,7 @@ def cli(
     ##
 
     writer = OutputWriter(
-        src_files,
-        output_file,
+        cast(BinaryIO, output_file),
         click_ctx=ctx,
         dry_run=dry_run,
         emit_header=header,
